\documentclass{report}

%    ****** TURN OFF HARDWARE TABS BEFORE EDITING THIS DOCUMENT ******
%
%   Should you ignore this admonition, tabs in the program code will
%   not be respected in the LaTeX-generated program document.
%   If that should occur, simply pass this file through
%   expand to replace the tabs with sequences of spaces.

%\usepackage{html}
\usepackage{drop}

%   This program is written using the NUWEB Literate Programming
%   tool:
%
%           http://sourceforge.net/projects/nuweb/
%
%   For information about Literate Programming, please visit the
%   site:   http://www.literateprogramming.com/

\setlength{\oddsidemargin}{0in}
\setlength{\evensidemargin}{0in}
\setlength{\topmargin}{0in}
\addtolength{\topmargin}{-\headheight}
\addtolength{\topmargin}{-\headsep}
\setlength{\textheight}{8.9in}
\setlength{\textwidth}{6.5in}
\setlength{\marginparwidth}{0.5in}
\setcounter{tocdepth}{4}
\setcounter{secnumdepth}{4}

%%%PDF%%%\usepackage[pdftex,pdftitle={Gardol},pdfauthor={John Walker},colorlinks=true,linkcolor=blue]{hyperref}

\title{{\bf Gardol} \\ {\em Adaptive Denial of Service Attack Mitigation}}
\date{March 2004}
\author{by
%%%PDF%%%\href{http://www.fourmilab.ch/}{
John Walker
%%%PDF%%%}
}

\begin{document}

\pagenumbering{roman}
\maketitle
\tableofcontents

\chapter{Introduction}
\pagenumbering{arabic}

\drop{D}{istributed} Denial of Service (DDoS) attacks are one of
a system administrator's worst nightmares.  A perfectly
functioning, highly secure, redundantly configured, carefully
firewalled system with adequate growth capacity of all kinds
can, in a matter of moments, be brought to its knees by being
bombarded by a huge number of apparently legitimate
simultaneous requests generated by by machines all over the
world. Driven into total overload, the Internet bandwidth
and/or server capacity saturated by the attackers, the attacked
site is rendered as inaccessible to legitimate users as if its
computer room were firebombed.  A company whose lifeblood is
online commerce can find its artery severed at the whim of
a malicious attacker.

DDoS attacks are, sadly, not the stuff of science fiction or a
worse-case scenario spun by security consultants peddling their
services.  They are, sadly, a fixture of today's Internet.  The
site operated by the author of this program has come under such
attacks several times since the year 1999, and has been under a
continuous, concerted attack generating up to half a million
bogus requests to its server since January 2004.  It is that
attack which motivated development of this program.

A Distributed Denial of Service Attack is {\em Distributed} because it
originates not from a single source (which could be easily identified
and blocked), but from a large collection of machines {\em distributed}
geographically.  None of these machines need, by itself, mount an
intensive attack upon the target.  It suffices that the collection of
simultaneously attacking machine, together, generate sufficient volume
to harm the intended victim.  In the attack against {\tt www.fourmilab.ch}
in early 2004, up to five thousand machines all around the world
simultaneously hit the site, and over a period of several weeks, more
than twenty thousand different machines (identified by IP address---some
may be the same physical machines appearing as different dynamically
assigned IP addresses) were observed as participating in the attack.
And the attack on Fourmilab was {\em minor} as DDoS attacks go; the
attack on {\tt sco.com} in January 2004 knocked that company's site
off the Web and forced them to change their domain name.

How can this happen?

\chapter{System Environment Parameters}

Set the following parameters to correspond to the system on
which you're installing the software.

\section{Directory where Perl is installed.}
@d Perl directory
@{/usr/bin/perl@}

\section{Program Version}
@d Version
@{0.1@}

\section{Release Date}
@d Release Date
@{March 2004@}

\section{Web Installation Directory}
@d Web Directory @{/ftp/webtools/gardol@}

\section{Web Log File Directory}
@d Web Log File Directory
@{/vitesse/server/logs/http@}

\section{Operating System Default Parameters} \label{defargs}

The following parameters specify operating system installation
specific path names and commands which are substituted
when an argument of ``{\tt -}'' is given for the corresponding
option on the command line.  This is purely a convenience to avoid
length command lines when doing production testing and may
be dispensed with in the interest of cleanliness if desired.

\subsection{Restart IP Filter Command}
When the IP Filter block list is updated, the following
command is used to cause IP Filter to load the updated list
and put it into effect.

@d Restart IP Filter Command
@{/etc/init.d/ipfboot reipf@}

\subsection{IP Filter Configuration File Template}

The following path name is used to read the IP Filter
configuration file template when the default operating
system configuration is delected.

@d IP Filter Configuration File Template
@{/etc/opt/ipf/ipf_conf_template.txt@}

\subsection{IP Filter Configuration File}

When the IP Filter configuration is updated, it is written to this
file when the default operating system configuration is
selected.

@d IP Filter Configuration File
@{/etc/opt/ipf/ipf.conf@}


\vbox{
\chapter{{\tt gardol.pl}: Attack Detection and Response}

}

\vbox{
\section{Attack Detection}

The following two sections of code are the heart of Gardol,
and where, as a system administrator using it to respond to
an attack, where you need to invest the most work.  From
the initial detection and analysis of the attack, you must
determine a unique signature, detectable by analysis of the
HTTP access log, which allows you to discriminate attacking
hosts from legitimate users by their pattern of access.  The
code which follows applies this test on individual log items.

For reasons of efficiency, the test is performed in two phases
(either of which may be void).  In the first, or ``quick reject''
phase, fields in the HTTP log item has been parsed out into the Perl
numeric capture variables but no further processing, such as computing
the Unix {\tt time()} value from the date and time in the log
item has been performed.  Any tests you can make at this phase
which exclude the log item as representating a potential attack
permit reducing the overhead of Gardol's monitoring the log.
Note that you can't use pattern matches in the quick reject
phase, as they would destroy the values in the capture variables,
but you shouldn't be doing pattern matches in a quick reject
in any case.  Records which can be excluded as belonging to
an attack are discarded by executing the
\verb+@@<Reject the current log item as benign@@>+ macro.

The following table lists the variables into which fields from the
HTTP log are decoded.  Fields which are present in the basic
CERN/NCSA ``Common Log Format'' are labeled as ``Common''
in the {\bf Log Format} column.  Fields which are present only
when reading an extended log format are identified by the
log format which provides them.  Each log format which appears
lower in the table is a superset of those which precede it.
The numeric capture variables (e.g.\ ``\verb+$5+'') are
available only in the quick reject phase.  Named variables
are defined before the full rejection phase is performed.  Items
with no entry in the {\bf Quick} column are not available in
that phase.  One detail: if the length of the HTTP response to
the requestor was zero, the \verb+$7+ variable in the quick
reject phase will be ``{\tt -}'', but the \verb+$length+
variable in the full reject phase will be set to 0.
\begin{center}
\begin{tabular}{|l|l|l|l|}
\hline
{\bf Quick} & {\bf Full} & {\bf Item} & {\bf Log Format} \\
\hline
 \verb+$1+  &   \verb+$ip+ & IP Address & Common  \\
 \verb+$2+  &   \verb+$ident+ & Identification & Common    \\
 \verb+$3+  &   \verb+$userid+ & User ID & Common    \\
 \verb+$4+  &   \verb+$time_date+ & Date and Time & Common    \\
 \verb+$5+  &   \verb+$request+ & HTTP Request & Common    \\
 \verb+$6+  &   \verb+$status+ & Status Code & Common    \\
 \verb+$7+  &   \verb+$length+ & Length of Response & Common    \\
 \verb+$8+  &   \verb+$referer+ & {\tt Referer} & Combined    \\
 \verb+$9+  &   \verb+$agent+ & {\tt User-Agent} & Combined    \\
 \verb+$10+ &   \verb+$cachecont+ & {\tt Cache-Control} & Forensic    \\
 \verb+$11+ &   \verb+$pragma+ & {\tt Pragma} & Forensic    \\
 \verb+$12+ &   \verb+$proxy+ & {\tt X-Forwarded-For} & Forensic   \\
 \hline
 \multicolumn{4}{|c|}{{\em Decoded Request Date and Time}} \\
 \hline
 &  \verb+$mday+ & Day of Month & Common  \\
 &  \verb+$mon+  & Month abbreviation (e.g. ``{\tt Mar}'' & Common  \\
 &  \verb+$year+ & Year & Common  \\
 &  \verb+$hour+ & Hour & Common  \\
 &  \verb+$minute+ & Minute & Common  \\
 &  \verb+$second+ & Second & Common  \\
 &  \verb+$timezone+ & Time zone offset & Common  \\
 &  \verb+$mindex+ & Month index $(0--11)$ & Common  \\
 &  \verb+$iso_date+ & ISO 8861 Date and Time & Common  \\
 &  \verb+$utime+ & Unix {\tt time()} value & Common  \\
 
\hline
\end{tabular}
\end{center}
}

\vbox{
\subsection{Quick reject non-attack HTTP log items}

Unless the attack we're under is completely unsustainable (so
intense it clogs our inbound pipe), once we've blocked most of
the currently-attacking IP addresses, most of the remaining
accesses in the HTTP log will be legitimate (non-attack-related).
We want to ignore these with as little overhead as possible, so
all tests which can be made from the raw items parsed from the
HTTP log item should be tested for here.

Obviously, you're going to need to customise this code to reject
records which aren't part of the particular attack on your
site and pass those which are.  {\em Important:} to reduce
overhead, these tests are performed directly on the numbered
patterm match capture variables parsed from the HTTP log item.
{\em Do not} use a pattern match in this code, as it will destroy
the values of these variables.  Pattern matches don't belong in
quick-reject code anyway.

@d Quick reject non-attack HTTP log items
@{
    #   Special status 573 indicates an attack hit pre-detected by the
    #   filter in Apache.
    if ($6 != 573) {
        #   But if an apparently attacking host sends a single packet
        #   which doesn't resemble an attack, remove it from the list
        #   of attackers.  Its future behaviour will determine whether
        #   we put it back on the list of miscreants.
        if (defined($hits{$1})) {
            @<Delete host from list of attackers@($1@)@>
            if ($verbose) {
                print(STDERR "--Purged by non-homepage hit $1\n");
            }
        }
        @<Reject the current log item as benign@>
    }

    #   Quick reject hits relayed by proxy servers
    if ($12 ne '-') {
        @<Reject the current log item as benign@>
    }
@}

To make quick and regular rejection code more readable, we define
a little macro to do the {\tt next} which causes the current log item
to be ignored.

@d Reject the current log item as benign
@{  next; @}
}

\vbox{
\subsection{Reject non-attack HTTP log items}

Where possible, detecting and rejection of non-attack log items
should be performed as a quick rejection in the previous section.
More complex rejection tests which require pattern matches,
access to the decoded time and date, etc. should be done
here, after the items parsed from the HTTP log item have been
assigned to their respective variables.

@d Reject non-attack HTTP log items
@{
    #   There are no items requiring non-quick rejection here.
    #   An example of a non-quick rejection test is:
    #
    #       if (!(($request =~ m-/earthview/cache/-) && ($status == 404))) {
    #           @<Reject the current log item as benign@>
    #       }
@}
}


\vbox{
\section{Main Program}
@o gardol.pl -i
@{#! @<Perl directory@>

    @<Documentation in POD format@>

    @<Global declarations@>
    
    @<Process command line options@>
    @<Validate option specifications@>
    
    if ($#ARGV != 1) {
        &print_command_line_help;
        exit(0);
    }
    
    #   HTTP Log file to monitor
    my $read_HTTP_log = $ARGV[0];
    
    #   IP Filter packet disposition log
    my $read_IP_Filter_log = $ARGV[1];
        
    @<Open log files to be monitored@>;
    
    @<Start monitoring for restart signal@>
    
    while (1) {
    
        @<Check for restart signal and reset log files if received@>
    
        @<Examine newly appended HTTP log items@>
        
        @<Examine newly appended IP Filter log items@>
        
        if ((time() - $lastupd) >= $updtime) {
            @<Update filter to block attacking hosts@>
        }
        if ($verbose) {
            print(STDERR "--Sleeping $sleepytime seconds.\n");
        }
        sleep($sleepytime);
    }
             
    @<Utility functions@>
@| $read_HTTP_log $read_IP_Filter_log @}
}

\vbox{
\subsection{Open log files to be monitored}

We monitor two log files: the HTTP log to track incoming requests
and detect sequences of requests which identify an attacking site,
and the IP Filter log to watch for continuing hits from IP addresses
we've already decided to ban.  (Monitoring the IP Filter log permits
timing out IP addresses which have ceased to hit us, which both reduces
the length of the list of blocked addresses and allows subsequent access
from a one-attacking floating IP address which has now been reassigned
to a benign user.)

@d Open log files to be monitored
@{
    open(FL, "<$read_HTTP_log") || die "Unable to open HTTP log file $read_HTTP_log";
    seek(FL,  0, 2);            # Seek to end of log file

    open(FLOG, "<$read_IP_Filter_log") || die "Cannot open IP Filter log $read_IP_Filter_log";
    seek(FLOG,  0, 2);          # Seek to end of IP Filter log file
  @}
}

\vbox{
\subsection{Start monitoring for restart signal}

We register to catch the ``{\tt HUP}'' signal which, when
received, causes the files we're monitoring to be closed, re-opened,
and re-seeked (re-sought?) to the end.  This allows monitoring
to be transferred when the log files are ``cycled'' without
killing and restarting the program.  Upon receiving the
signal, we simply set the variable \verb+$restart_signal_received+
which causes the closing and re-opening of the files the next
time we go around the main loop.  This avoids any problems due to
lack of re-entrancy in the underlying system or by snatching file
descriptors out from under code in this program.

@d Start monitoring for restart signal
@{
    $SIG{HUP} = sub {
        $restart_signal_received++;
#print("Restart! ($restart_signal_received)\n");
    };
@}
}


\vbox{
\subsection{Check for restart signal and reset log files if received}

If a restart signal has been received since the last time we went around
the main loop, close and re-open the log files we're monitoring so
if they've been cycled to new files, we'll commence monitoring them.

@d Check for restart signal and reset log files if received
@{
    if ($restart_signal_received) {
        close(FL);
        close(FLOG);
        @<Open log files to be monitored@>
        $restart_signal_received = 0;
        if ($verbose) {
            print(STDERR "Log files re-opened.\n");
        }
    }
@}
}


\vbox{
\subsection{Process HTTP log items}

Read all HTTP log items appended since the last time around
the main loop.  Each is parsed, then tested to see if it
is part of the attack.  If so, we add its source to the
list of currently-attacking hosts and record the time of
the first attack (should it so be) and
most recent attack from this host.

@d Examine newly appended HTTP log items
@{
    while ($l = <FL>) {
    
        #   Parse request record
        if ($l !~ m/$parseHTTPlog/) {
            #   Can't parse log item.  Ignore it.
            if ($verbose) {
                print(STDERR "!! Cannot parse HTTP log item: $l");
            }
            @<Reject the current log item as benign@>
        }

        @<Quick reject non-attack HTTP log items@>

        @<Parse HTTP log item into variables@>
        
        @<Reject non-attack HTTP log items@>

        @<Add IP address to list of attacking hosts@>
    }
@}
}

\vbox{
\subsubsection{Parse HTTP log item into variables}

If quick rejection has not excluded this HTTP log item as benign,
we proceed to store the fields parsed from it into
individual variables and further parse the extracted
time and date field into its components, which are then
re-assembled into an ISO-8601 date and time and
a Unix {\tt time()} value for arithmetical comparisons.

@d Parse HTTP log item into variables
@{
    $ip = $1;
    $ident = $2;
    $userid = $3;
    $time_date = $4;
    $request = $5;
    $status = $6;
    $length = $7;
    $referer = $8;
    $agent = $9;
    $cachecont = $10;
    $pragma = $11;
    $proxy = $12;
    if ($length eq '-') {
        $length = 0;
    }

    #   Parse date and time field
    $time_date =~ m-(\d+)/(\w+)/(\d+):(\d+):(\d+):(\d+)\s([\+\-]\d+)$-;
    $mday = $1;
    $mon = $2;
    $year = $3;
    $hour = $4;
    $minute = $5;
    $second = $6;
    $timezone = $7;
    $mindex = $mnames{$mon};
    $iso_date = sprintf("%04d-%02d-%02d %02d:%02d:%02d", $year, $mindex, $mday,
                            $hour, $minute, $second);
    $utime = timelocal($second, $minute, $hour, $mday, $mindex - 1, $year);
    
#print("$mday,$iso_date,$year,$hour,$minute,$second,$timezone\n");
#print ("$ip,$ident,$userid,$time_date,$request,$status,$length,$cachecont,$pragma,$proxy\n");
@| $ip $ident $userid $time_date $request $status $length $referer $agent $cachecont $pragma $proxy
   $time_date $mday $mon $year $hour $minute $second $timezone $mindex $iso_date $utime @}
}

\vbox{
\subsubsection{Add IP address to list of attacking hosts}

This HTTP log item has been deemed an attack (or attack
candidate, subject to further scrutiny when it's time to
decide which IP addresses should be blocked).  Increment
the number of hits from the IP address, save the time of
this hit and, if this is the first hit from this address,
save the request time as the time of the first hit as well.

@d Add IP address to list of attacking hosts
@{
    $hits{$ip}++;
    if ($verbose) {
        print(STDERR "--($hits{$ip}) $l");
    }
    if (!defined($ufirst{$ip})) {
        $ufirst{$ip} = $utime;
    }
    $ulast{$ip} = $utime;
@| $hits $ufirst $ulast %hits %ufirst %ulast @}
}

\vbox{
\subsection{Process IP Filter log items}

Once a host has been deemed an attacker and its IP address placed
in the IP Filter block list, we will no longer see hits from it
in the HTTP log.  Nonetheless, the host may very well continue to
hit us, even without the satisfaction of a reply (this is the
case for the brute denial attack which motivated the development
of this program).  If we relied exclusively on the HTTP log, once
we denied access to the IP address, its timeout would start ticking
down and, without any new hit appearing in the log, would expire.
The host would then be granted access again, would resume the attack,
be blocked once more, and so on.

To avoid this silly cycle, we examine records appended to the
IP Filter log since the last cycle.  This log is written with the
{\tt ipmon} program which comes with IP Filter; the file name
containing this log is the second command line argument on
the call to this program.  For each new IP Filter log item, we parse
the IP address, date and time, and disposition of the packet.  If
this is a packet blocked originating from an IP address we've
deemed an attacker, reset the timeout to the time the blocked packet
was received and increment the count of hits from this host.  This
will guarantee that a blocked host will continue to remain blocked
as long as it continues to hit, even though the HTTP server never sees
the blocked packets.

@d Examine newly appended IP Filter log items
@{
    while ($l = <FLOG>) {
        if ($l =~ m/$parse_IP_Filter/) {
            $utime = timelocal($6, $5, $4, $1, $2 - 1, $3);
            $bp = $8;
            $ip = $9;
            if (($bp eq 'b') && (defined($ufirst{$ip}))) {
                $ulast{$ip} = $utime;
                $hits{$ip}++;
                if ($verbose) {
                    print(STDERR "-+($hits{$ip}) $l");
                }
            }
        }
    }
@}
}


\vbox{
\subsection{Update filter to block attacking hosts}

The filter file (\verb+$filtfile+) consists of a sequence of
configuration commands for IP Filter.  It is assembled by
transcribing a fixed template file (\verb+$templatefile+)
until a sentinel (\verb+$templatesentinel+) is encountered,
interpolating the list of hosts to be blocked and any commands
from the {\tt --blacklist} if any, then appending the balance
of the template to complete the filter configuration file.
If a \verb+$reload_IP_Filter+ command has been specified, it
is executed to notify the filter to put the updated rules into
effect.

@d Update filter to block attacking hosts
@{
    my $now = time();

    open(TE, "<$templatefile") || die "Unable to open template file $templatefile";
    open(OF, ">$filtfile") || die "Unable to create IP Filter configuration file $filtfile";

    @<Transcribe template to IP Filter configuration file@>

    @<Generate list of IP addresses to be blocked@>
    
    @<Add blacklisted IP addresses to block list@>

    @<Transcribe template to IP Filter configuration file@>

    @<Remove timed out hosts from list of attackers@>

    close(OF);
    if ($verbose) {
        printf(STDERR "--Blocked $nblock hosts.\n");
    }
    $lastupd = time();

    if ($reload_IP_Filter ne '') {
        system("$reload_IP_Filter\n");
    }
@}
}

\vbox{
\subsubsection{Transcribe template to IP Filter configuration file}

Copy the IP Filter configuration template to the configuration
file about to be put into service, stopping when we encounter
the \verb+$templatesentinel+.

@d Transcribe template to IP Filter configuration file
@{
    while (<TE>) {
        if ($_ =~ m/^$templatesentinel/) {
            last;
        }
        print(OF);
    }
@}
}

\vbox{
\subsubsection{Generate list of IP addresses to be blocked}

Walk through the list of hosts suspected to be attacking.  Any host
whose last hit is longer than the timeout interval is skipped and
marked for deletion from the list of attacking hosts.  Hosts
whose last hit was within the timeout interval are tested against
the denial criterion (usually based on the number of hits or
mean hit rate) and, if deemed an attacker, are listed in
an IP Filter {\tt block} rule to bar access from that
IP address.

@d Generate list of IP addresses to be blocked
@{
    my @@toh;
    my $nblock = 0;
    my $k;

    foreach $k (sort keys %ulast) {
        if (($now - $ulast{$k}) >= $timeout) {
            push(@@toh, $k);
        } else {
            if ($hits{$k} >= $minhits) {
                my $mina = int(($now - $ufirst{$k}) / 60);
                print(OF "block in log quick from $k to any  # $mina min. $hits{$k} hits\n");
                $nblock++;
            }
        }
    }
@}
}

\vbox{
\subsubsection{Add blacklisted IP addresses to block list}

If a {\tt --blacklist} file is specified and exists, transcribe
the IP addresses from the blacklist to the list of hosts to be
blocked by the filter.  IP addresses to be blacklisted appear
one per line.  Comments which follow a \verb+#+ and blank lines
are ignored.  The actual blacklist command is not validated--if it's
invalid, it's up to IP Filter to issue a warning and ignore it.
Passing the blacklist specification through directly permits it
to be a general expression, possibly including a netmask or
other modifier.

Lines in the blacklist file which begin with ``{\tt +}'' (after any
leading white space is discarded) are transcribed to the
blacklist file literally, dropping the plus sign.  This permits
specification of more complicated rules than a simple IP-based
block to all destinations.  Again, these statements are not
syntax checked, so errors may lead to warnings or more dire
consequences when IP Filter attempts to digest them.

@d Add blacklisted IP addresses to block list
@{
    if (($blacklistfile ne '') && (-f $blacklistfile)) {
        my $b;
        
        if (open(BL, "<$blacklistfile")) {
            while ($b = <BL>) {
                $b =~ s/#.*$//;
                $b =~ s/^\s*//;
                $b =~ s/\s+$//;
                if ($b ne '') {
                    if ($b =~ s/^\+//) {
                        print(OF "$b  # Transcribed from --blacklist $blacklistfile\n");
                    } else {
                        print(OF "block in log quick from $b to any  # Included from --blacklist $blacklistfile\n");
                    }
                }
            }
            close(BL);
        }
    }
@}
}

\vbox{
\subsubsection{Remove timed out hosts from list of attackers}

As we scan the list of attacking hosts, any which haven't hit
us since the \verb+$timeout+ interval are added to the
\verb+@@toh+ (timed-out hosts) lists.  We can't delete
them on the fly because that disrupts the {\tt foreach}
statement we use to scan the list of hosts.  So$\ldots$we
chew through the \verb+@@toh+ list here and forget everything
we knew about the hosts it cites.

@d Remove timed out hosts from list of attackers
@{
    foreach $k (@@toh) {
        @<Delete host from list of attackers@($k@)@>
        if ($verbose) {
            print(STDERR "--Timeout purged $k\n");
        }
    }
@}
}

\vbox{
\subsubsection{Delete host from list of attackers}

This little macro purges the host with the IP address of
the argument from all our tables of attacking hosts.

@d Delete host from list of attackers
@{
    delete($ulast{@1});
    delete($ufirst{@1});
    delete($hits{@1});
@}
}

\vbox{
\subsection{Template}


@d Template
@{
@}
}

\section{Global declarations}

@d Global declarations
@{
    @<Perl language modes@>
    
    use Time::Local;
        
    @<Log file parsing patterns@>

    @<Default parameter settings@>
            
    @<Global variables@>

@}

\section{Perl language modes}

@d Perl language modes
@{
    require 5;
    use strict;
@}

\vbox{
\subsection{Default parameter settings}

The following variables contain parameters which control
the operation of the program.  All of these are defaults
which can be overridden by command line options.

@d Default parameter settings
@{

    #   HTTP transfer log parsing pattern
    my $parseHTTPlog = $parse_forensic;

    #   How long to sleep between scans of new log items
    my $sleepytime = 10;
        
    #   Frequency of update dumps in seconds
    my $updtime = 5 * 60;
    
    #   Time out hosts after this many seconds without a hit
    my $timeout = 30 * 60;
    
    #   Minimum hits from IP address to deem an attacker
    my $minhits = 2;
    
    #   Where to read IP Filter configuration file template
    my $templatefile = 'ipf_conf_template.txt';
    
    #   Sentinel marking where rules are interpolated in ipf.conf file
    my $templatesentinel = '@<IP Filter Template Sentinel@>';
    
    #   Where to read explicit list of IP addresses to blacklist
    my $blacklistfile = '';
    
    #   Where to write IP Filter configuration file
    my $filtfile = 'ipf.conf';
    
    #   Command to load new rule set into IP Filter
    my $reload_IP_Filter = '';

@| $parseHTTPlog $sleepytime $updtime $timeout $minhits $templatefile $templatesentinel
   $blacklistfile $filtfile $reload_IP_Filter @}
}

\vbox{
\subsection{Global variables}

The following variables are global to the entire
{\tt main} program.  We could make many of
these more local, but the entire program is sufficiently
short and straightforward we'd probably only end up
obfuscating things in the interest of ``purity.''

@d Global variables
@{
    #   Processing arguments and options
    
    my $verbose = 0;            # Verbose output for debugging
    
    #   Handy constants
    
    my %mnames = split(/,/, "Jan,1,Feb,2,Mar,3,Apr,4,May,5,Jun,6,Jul,7,Aug,8,Sep,9,Oct,10,Nov,11,Dec,12");
    
    #   Utility variables

    my $lastupd = 0;            # Time of last update dump
    my $restart_signal_received = 0; # Nonzero if HUP restart signal received
    
    my ($l, $ip, $ident, $userid, $time_date, $request, $status, $length, $referer,
           $agent, $cachecont, $pragma, $proxy, $time_date, $mday, $mon, $year,
           $hour, $minute, $second, $timezone, $mindex, $iso_date, $utime, $bp);
           
    my (%hits, %ufirst, %ulast);
@| $verbose $mnames $lastupd $restart_signal_received @}
}

\vbox{
\subsection{Log file parsing patterns}

The following declarations define the regular expression patterns used to parse
the different log file formats we read.

@d Log file parsing patterns
@{
    @<Common log parsing pattern@>
    @<Combined log parsing pattern@>
    @<Forensic log parsing pattern@>
    
    @<IP Filter log parsing pattern@>
@}
}

\vbox{
\subsubsection{Common log parsing pattern}

The following expression parses an Apache HTTP log in the CERN/NCSA
``common log file'' format, which is defined in the standard
Apache {\tt httpd.conf} file as follows:

\begin{verbatim}
LogFormat "%h %l %u %t \"%r\" %>s %b" common
\end{verbatim}

@d Common log parsing pattern
@{
    my $parse_common = qr/^(\d+\.\d+\.\d+\.\d+)\s+(\S+)\s+(\S+)\s+\[(.*)\]\s+"(.*)"\s+(\d+)\s+([\-\d]+)"/;
@| $parse_common @}
}

\vbox{
\subsubsection{Combined log parsing pattern}

The following expression parses an Apache HTTP log in the
``combined log file'' format, which contains the same 8 initial
fields as the common log file format, plus quoted fields which
specify the ``{\tt Referer}'' and ``{\tt User-Agent}'' from the
HTTP request, or ``{\tt -}'' if these items did not appear
in the request.

\begin{verbatim}
LogFormat "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\"" combined
\end{verbatim}

@d Combined log parsing pattern
@{
    my $parse_combined = qr/^(\d+\.\d+\.\d+\.\d+)\s+(\S+)\s+(\S+)\s+\[(.*)\]\s+"(.*)"\s+(\d+)\s+([\-\d]+)"/;
@| $parse_combined @}
}

\vbox{
\subsubsection{Forensic log parsing pattern}

The following expression parses an Apache HTTP log in the special
``forensic log'' format developed to diagnose and respond to the
2004 attack against {\bf www.fourmilab.ch}.  This format consists
of the ``{\tt combined}'' log format with quoted fields appended
which contain the ``{\tt Cache-Control}'', ``{\tt Pragma}'',
and ``{\tt X-Forwarded-For}'' HTTP request header field specifications
(or ``{\tt -}'' if the field is absent).

To create a log in forensic format, add the following to your
Apache {\tt httpd.conf} file, adjusting the location where the
log should be written in the {\tt CustomLog} statement as
appropriate.

\begin{verbatim}
LogFormat "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\" \"%{Cache-Control}i\" \"%{Pragma}i\" \"%{X-Forwarded-For}i\"" forensic
CustomLog /files/server/logs/http/forensic_log forensic
\end{verbatim}

@d Forensic log parsing pattern
@{
    my $parse_forensic = qr/^(\d+\.\d+\.\d+\.\d+)\s+(\S+)\s+(\S+)\s+\[(.*)\]\s+"(.*)"\s+(\d+)\s+([\-\d]+)\s"((?:[^"]|"")*)"\s"((?:[^"]|"")*)"\s"((?:[^"]|"")*)"\s"((?:[^"]|"")*)"\s"((?:[^"]|"")*)"/;
@| $parse_forensic @}
}

\vbox{
\subsubsection{IP Filter log parsing pattern}

We track records appended to the IP Filter log in order to reset the
inactivity timeout whenever a packet from an attacking host is blocked
by the filter rules we've put into effect.  These records look like
the following:

\begin{verbatim}
05/03/2004 16:32:51.670599 hme0 @@0:909 b 219.77.161.203,45931 -> 193.8.230.138,80 PR tcp len 20 48 -S IN
\end{verbatim}

and we're interested in parsing the date and time, whether the packet was blocked
(``{\tt b}'') or passed (``{\tt p}''), and the source IP address.

@d IP Filter log parsing pattern
@{
    my $parse_IP_Filter = qr-^(\d+)/(\d+)/(\d+)\s+(\d+):(\d+):(\d+)\.(\d+)\s+\w+\s+\@@\d+:\d+\s+(\w)\s+(\d+\.\d+\.\d+\.\d+),-;
@| $parse_IP_Filter @}
}


\vbox{
\subsection{Process command line options}

We use the {\tt Getopt::Long} module to process command line
options.

@d Process command line options
@{
    use Getopt::Long;
        
    GetOptions(
                'blacklist=s' => \$blacklistfile,
                'copyright' => sub { print("This program is in the public domain.\n"); exit(0); },
                'help' => sub { &print_command_line_help; exit(0); },
                'ipfconf=s' => \$filtfile,
                'ipfsentinel=s' => \$templatesentinel,
                'ipftemplate=s' => \$templatefile,
                'ipfupdate=s' => \$reload_IP_Filter,
                'minhits=i' => \$minhits,
                'polltime=i' => \$sleepytime,
                'timeout=i' => \$timeout,
                'updtime=i' => \$updtime,
                'verbose' => \$verbose,
                'version' => sub { print("Version @<Version@>, @<Release Date@>\n"); exit(0); }
              );
              
@}
}

\vbox{
\subsection{Validate option specfications}

Validate the option specifications before we begin processing.
Pre-checking them avoids ugly pratfalls with the input half-processed.

@d Validate option specifications
@{

    {
        my $ok = 1;
        
        if (($blacklistfile ne '') && (!(-f $blacklistfile))) {
            print(STDERR "Blacklist file $blacklistfile does not exist.\n");
            $ok = 0;
        }
        
        if ($reload_IP_Filter eq '-') {
            $reload_IP_Filter = '@<Restart IP Filter Command@>';
        }
        
        if ($templatesentinel eq '-') {
            $templatesentinel = '@<IP Filter Configuration File Template@>';
        }
        
        if ($templatefile eq '-') {
            $templatefile = '@<IP Filter Configuration File@>';
        }
                
        if (!$ok) {
            die("Invalid option specification(s)");
        }
    }
@}
}


\vbox{
\section{Utility Functions}

The following utility functions are defined in the main program
context to handle matters such as command line processing.

@d Utility functions
@{
    @<Print command line help information@>
@}
}


\vbox{
\subsection{Print command line help information}

@d Print command line help information
@{
    sub print_command_line_help {
        print << "EOD";
Usage: gardol.pl [ options ] HTTP_log_to_monitor IP_Filter_log_to_monitor
       Options:
             --blacklist f   Add IP addresses in blacklist file f to block list
             --copyright     Print copyright information
             --help          Print this message
             --ipfconf f     Write IP Filter configuration to file f
             --ipfsentinel s Interpolate host blocking commands at this sentinel in IP Filter template
             --ipftemplate f Read IP Filter configuration template from file f
             --ipfupdate c   Use shell command c to update IP Filter configuration
             --minhits n     Require n consecutive attack packets before blocking host
             --polltime n    Poll log files for new items every n seconds
             --timeout n     Timeout and unblock hosts after n seconds of inactivity
             --updtime n     Update list of blocked sites every n seconds
             --verbose       Print verbose debugging information
             --version       Print version number
Version @<Version@>, @<Release Date@>
EOD
   }
@}
}

\vbox{
\section{Documentation in POD format}

@d Documentation in POD format
@{
=head1 NAME

gardol - Adaptive Denial of Service Attack Mitigation

=head1 SYNOPSIS

B<gardol.pl>
[I<options>]
I<HTTP_log_to_monitor>
I<IP_Filter_log_to_monitor>

=head1 DESCRIPTION

B<gardol> is a Perl program which monitors an HTTP log file to
detect accesses which identify IP addresses participating in a
Distributed Denial of Service (DDoS) attack and, in conjunction
with the
S<B<IP Filter>>
package, dynamically block them from attacking a server.

@<Options documentation@>

=head1 VERSION

This is B<gardol> version @<Version@>, released @<Release Date@>.
The current version of this program is always posted at
http://www.fourmilab.ch/webtools/gardol/.

=head1 AUTHOR

John Walker
(http://www.fourmilab.ch/)

=head1 BUGS

Please report any bugs to bugs@@fourmilab.ch.

=head1 SEE ALSO

B<IP Filter> (http://coombs.anu.edu.au/~avalon/),
B<nuweb> (http://nuweb.sourceforge.net/),
S<Literate Programming> (http://www.literateprogramming.com/).

=head1 COPYRIGHT

This program is in the public domain.

=cut
@}
}

\vbox{
\subsection{Options}

Here we document the command-line options.

@d Options documentation
@{
=head1 OPTIONS

All options may be abbreviated to their shortest
unambiguous prefix.

=over 5
@}
}

\vbox{
\subsubsection{{\tt --blacklist} {\em filename}}
@d Options documentation
@{
=item B<--blacklist> I<filename>

Whenever the list of IP addresses to be blocked is updated, the
IP addresses listed in the specified I<filename> (one per line, with
blank lines and any text following a "#" character ignored) will be
added to the list of IP addresses to be blocked.
@}
}

\vbox{
\subsubsection{{\tt --copyright}}
@d Options documentation
@{
=item B<--copyright>

Display copyright information.
@}
}

\vbox{
\subsubsection{{\tt --help}}
@d Options documentation
@{
=item B<--help>

Display how to call information.

@}
}

\vbox{
\subsubsection{{\tt --ipfconf} {\em filename}}
@d Options documentation
@{
=item B<--ipfconf> I<filename>

The IP Filter configuration file containing the list of IP addresses
to be blocked is written to the specified I<filename>.
The default configuration file name is C<B<ipf.conf>> in
the current directory.  If a I<filename> of "B<->" is specified,
the installation default template location of
C<B<@<IP Filter Configuration File@>>> is specified.

@}
}

\vbox{
\subsubsection{{\tt --ipfsentinel} {\em string}}
@d Options documentation
@{
=item B<--ipfsentinel> I<string>

The list of IP addresses to be blocked is interpolated into the
B<--ipftemplate> file when a line containing the I<string> is
read.  The default sentinel is:

S<C<@<IP Filter Template Sentinel@>>>

@}
}

\vbox{
\subsubsection{{\tt --ipftemplate} {\em filename}}
@d Options documentation
@{
=item B<--ipftemplate> I<filename>

The specified I<filename> is used as the template for the IP
Filter configuration file.  Lines of the template are copied
to the configuration file, with the list of attacking IP
addresses to be blocked interpolated at the specified
B<--ipfsentinel>.
The default template file name is C<B<ipf_conf_template.txt>>
in the current directory.  If a I<filename> of "B<->" is specified,
the installation default template location of
C<B<@<IP Filter Configuration File Template@>>> is specified.

@}
}

\vbox{
\subsubsection{{\tt --ipfupdate} {\em command}}
@d Options documentation
@{
=item B<--ipfupdate> I<command>

The specified I<command> will be executed to inform
IP Filter when changes are made in the configuration
file, instructing it to load the new rules.  The default
command is blank, which disables reloading of the filter
rules.  Specifying a I<command> of "B<->" sets the reload command
to the IP Filter installation default of
C<B<@<Restart IP Filter Command@>>>.

@}
}

\vbox{
\subsubsection{{\tt --minhits} {\em n}}
@d Options documentation
@{
=item B<--minhits> I<n>

A total of I<n> consecutive attack packets (default 2) from a
given host will be required before its IP address is added to
the block list..

@}
}

\vbox{
\subsubsection{{\tt --polltime} {\em t}}
@d Options documentation
@{
=item B<--polltime> I<t>

The HTTP and IP Filter log files will be checked for newly appended
records every I<t> (default 10) seconds.  Log records are read and
processed until the end of the log file is encountered, whereupon
B<gardol> sleeps for I<t> seconds before checking whether additional
items have been added.

@}
}

\vbox{
\subsubsection{{\tt --timeout} {\em t}}
@d Options documentation
@{
=item B<--timeout> I<t>

Hosts from which neither an attack packet nor a packet blocked by
IP Filter has been received in I<t> seconds (default 1800 seconds--30
minutes) are timed out and access from their IP addresses re-enabled.
This prevents floating IP addresses from remaining blocked once an
attacker ceases to hit from them.

@}
}

\vbox{
\subsubsection{{\tt --updtime} {\em t}}
@d Options documentation
@{
=item B<--updtime> I<t>

The IP Filter configuration file is updated, adding host newly added
to the block list and removing those which have timed out, every
I<t> seconds (default 300--five minutes).  The
B<--updtime> should be sent sufficiently short to respond to
the changing population of attacking hosts in a timely
manner, but not so short as to create too much overhead updating
and reloading the filter configuration.

@}
}

\vbox{
\subsubsection{{\tt --verbose}}
@d Options documentation
@{
=item B<--verbose>

Generate verbose output to indicate what's going on.

@}
}

\vbox{
\subsubsection{{\tt --version}}
@d Options documentation
@{
=item B<--version>

Display version number.

=back
@}
}

\vbox{
\chapter{{\tt ipf\_conf\_template.txt}: IP Filter Configuration File Template}

@d IP Filter Template Sentinel
@{#### INSERT ATTACK_DENIAL RULES HERE ####@}

@o ipf_conf_template.txt -i
@{
## ipf.conf - config file for ipfilter
##
## pass all local traffic
pass in quick on lo0 all
pass out quick on lo0 all

## By default, pass all traffic (this is appropriate for a machine
## already situated behind a firewall).

pass in on hme0 all
pass out on hme0 all

## Ephemeral rules incorporated by Gardol

@<IP Filter Template Sentinel@>

## End ephemeral rules incorporated by Gardol

##
## end of ipfilter ruleset
@}
}

\vbox{
\chapter{{\tt index.html}: Main Web Page}

This is the main Web page for Gardol.  It contains the
user documentation and the initial request form.

@o index.html -i
@{

@<HTML header section@>

<body bgcolor="#FFFFFF" >

<center>
<h1><img src="figures/gardol.png" width="512" height="214" alt="GARDOL"></h1>
<h2>Adaptive Denial of Service Attack Mitigation</h2>
<h3>by <a href="/">John Walker</a></h3>
</center>

        
&nbsp;<p>


<div class="bodycopy">

@<Introductory text@>

@<Questions to answer@>

<h2>Download <cite>Gardol</cite>: &nbsp; <a href="gardol.tar.gz"><tt>gardol.tar.gz</tt></a> source archive</h2>
<h2>Read <cite>Gardol</cite> <a href="gardol.pdf">source code</a></h2>

<p>
<hr>

<h3><a href="/">Fourmilab Home Page</a></h3>

<address>
by <a href="/">John Walker</a><br>
@<Release Date@>
</address>
<center>
<em>This document is in the public domain.</em>
<br>&nbsp;
</center>

</div>

</body>
</html>
@}
}

\vbox{
\section{HTML Header Section}

@d HTML header section
@{
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html lang="en">
<head>
<title>Gardol: Adaptive Denial of Service Attack Mitigation</title>
<style type="text/css">
    DIV.bodycopy {
        margin-left: 10%;
        margin-right: 10%
    }
</style>
<XXbase href="http://www.fourmilab.ch/" >

<meta name="keywords" content="gardol, denial, service, attack, mitigation, tool, response">
<meta name="description" content="Gardol: Adaptive Denial of Service Attack Mitigation">
<meta name="author" content="John Walker">
<meta name="robots" content="index">

</head>
@}
}

\vbox{
\section{Introductory Text}

@d Introductory text
@{
@}
}


\vbox{
\section{Questions to Answer}

@d Questions to answer
@{
<h2>Questions to Answer</h2>

When an apparent distributed denial of service attack hits your
site, it's easy to panic, especially if it effectively knocks
your site off the Web and the boss is running around like a
beheaded chicken urging you to "do something".  This is
even more severe at sites like Fourmilab, where I have to play
both the rle of the chicken and the beleaguered system
administrator simultaneously.  Still, before you can ``do anything''
that's likely to improve the situation, you need to carefully determine
the facts of the matter, which process you can begin by answering the
following questions.

<h3>Are we really under attack?</h3>

You may have decided your site is under attack because you've observed
a large bulge in the hit rate on your server, seen your inbound
and/or outbound network bandwidth hit the peg, or watched the CPU
load on your server reach saturation.

<table align="right">
<tr><td align="center">
<img src="figures/2004-01.png" width="209" height="200" alt="Web access log: January 2004">
<br>
<small><em>
An attack on Fourmilab.
</em></small>
</table>

The first question to ask yourself is whether these symptoms are
actually the result of a DDoS attack or something else
entirely.  Certainly, all of these things are characteristic
of a DDoS attack, but they do not, by themselves, indicate an
attack is actually underway.  The chart at the right shows the onset
of the
<a href="http://www.fourmilab.ch/documents/security/incidents/ddos_2004-01/" target="_blank">DDoS
attack against Fourmilab</a>
in January 2004.  Prior to
the attack, the hits per day had been at the typical level of
500,000 to 650,000 per day.  As the attack began, hits per day
increased to over a million per day and stayed at that level.
Analysis of the request log showed the additional hits were
completely stylised requests for nothing but the home page,
originating from thousands of different IP addresses all over
the world, with each requestor repeating at a more or less
constant rate, and several hundred new IP addresses "recruited"
into the attacking population each hour.  The attack requests were
easily distinguished from legitimate requests for the home page
because they contained request header variables which no browser
was known to send.  In this case, an unexpected bulge in request rate
did indeed indicate an attack.  But this is not always the case.

<br clear="right">

<p>

Consider the access graph at the left, covering a week in April 2003.  Once
again, requests were running at the typical rate with the usual slowdown
during the week-end when, on the 13th of April, the hit rate exploded
to more than double the usual average rate.

<table align="left">
<tr><td align="center">
<img src="figures/2003-04.png" width="152" height="199" alt="Web access log: April 2003">
<br>
<small><em>
Not an attack.
</em></small>
</table>

This hit the outbound bandwidth hard--its 2 Mb/sec capacity, which is about twice
the average bandwidth requirement, saturated, and response time went from the
typical 100-500 milliseconds to in excess of five seconds for many
requests.  The hits just kept on coming at this rate for the first hours
of the 14th as well, at which time they finally began to taper off.  An
attack?  Nope.  Examination of the request log immediately revealed what
was going on.  One of the documents on out site had been mentioned on one
of those sites frequented by fat people who cannot spell, who click like
Skinner-pigeons on any link posted there, in order to download, but of course
never read, the contents before posting their erudite "arguements" against
its (unread) content and disdain for the author of its (unread) words.
You can usually identify the source of this kind of traffic bulge from
the "<tt>Referer</tt>" field in your HTTP log.  (If you don't use the
"combined" log file format which includes this field along with the
"<tt>User-Agent</tt>" which identifies the browser, here is an excellent
reason for starting to do so.)  In this case, most of the initial hits
to the document in question has a <tt>Referer</tt> pointing to a posting
on a site for iconoclasts who all think alike, and a visit there confirmed
that as the source of the hits.  One could then breathe a sigh of
relief, "this shall soon pass", and indeed, once the posting scrolled
off the home page of the site, the hit rate returned to normal.  A few
minutes of research put an end to the panic over a potential attack
and pointed to patience as the proper prescription.

<br clear="left">

<p>

One month earlier, in March 2003, after a typical week and
week-end, the request rate started climbing from its typical rate
to higher and higher levels on each successive day, reaching

<table align="right">
<tr><td align="center">
<img src="figures/2003-03.png" width="209" height="199" alt="Web access log: March 2003">
<br>
<small><em>
An attack, but not on Fourmilab.
</em></small>
</table>

more than a million hits a day by the 21st.  These hits originated
from all over the world, and were primarily requests for images
from
<cite><a href="/earthview/" target="_blank">Earth and Moon Viewer</a></cite>
which, requiring substantial CPU time on the server, vastly exceeded
the server's CPU capacity, sending its load average, which typically
peaks around 3.5 to 4, to more than 400 processes waiting to
run.  The referer field in the first hit of these requests were
mostly blank or from search engines.  Was Fourmilab under attack?
No--<em>Iraq was!</em>  The outbreak of hostilities in Iraq causes
tens of thousands of people all over the world, having seen satellite
images of Iraq on television, to run to their computers to look
for a source, whereupon Earth and Moon Viewer popped out at the
top of the search.  I suspect most of these visitors didn't realise
that the images they were requesting were generated from a
<em>static imagery database</em>, so no matter how far they zoomed
in, they wouldn't be able see the tanks roll toward Baghdad.

<br clear="right">

<h3>What is the kind of attack?</h3>

<h4>DDoS Attacks of the First Kind: Nuisance Level</h4>

<h4>DDoS Attacks of the Second Kind: Outbound Bandwidth / Server Saturation</h4>

<h4>DDoS Attacks of the Third Kind: Inbound Bandwidth Saturation</h4>

<h3>Where is it coming from?</h3>

<h4>What is the profile of the attacking hosts?</h4>

<h4>Is the attacking population static or dynamic?</h4>

<h4>What is the trend of the attack?</h4>

<h3>What are the attackers sending?</h3>

<h4>Is our inbound bandwidth being saturated?</h4>

<h4>Is our outbound bandwidth being saturated?</h4>

<h4>Is our server being overloaded?</h4>

<h4>Are the attack requests stereotyped?</h4>

<h4>Can attack requests be distinguished from legitimate ones?</h4>

<h4>What is the request trajectory of an attacker?</h4>

<h3>When did the attack start?</h3>

<h4>Who were the first identifiable attackers?</h4>
@}
}

\vbox{
\chapter{{\tt badbot.pl}: BadBot Attack Simulator}
\begin{verse}
%%%PDF%%%\href{http://phobos.apple.com/WebObjects/MZStore.woa/wa/viewAlbum?selectedItemId=165575&amp;amp;playListId=165606}{
    Bad bot, bad bot!\\
    How 'ya gonna fight?\\
    When somebody aims it,\\
    At your site.
%%%PDF%%%}
\end{verse}

In the immortal words of Kelvin R. Throop, ``If you haven't tried it, it
doesn't work.''  {\bf BadBot} allows you to simulate distributed
denial of service attacks against your site, allowing you to debug attack
detection code in {\bf Gardol}.  Running {\bf Gardol} nonprivileged, simply
monitoring the HTTP log file and writing a filter list in a private
directory, allows you to test detection of simulated attacks mounted
with {\bf BadBot} without disrupting the operation of your site in any
way.

{\bf BadBot} is controlled by a sequence of commands (they
look like options, but they're really imperative commands
processed in sequence) on its command line.  You can submit
{\tt GET}, {\tt HEAD}, and {\tt POST} requests to Web servers, wait
for a specified number of seconds, and loop repeating a sequence
of commands until the program is killed.  The {\tt --help} command
prints a list of all available commands.  For example, to reproduce
the every four minute hit on the home page attack which struck
Fourmilab in early 2004, the following command will suffice.

\begin{verbatim}
    perl badbot.pl --set Cache-control=no-cache \
                   --set Referer=- \
                   --agent '' \
                   --get http://www.victim.org/ --wait 240 --loop
\end{verbatim}

}

\vbox{
\section{Main program}
@o badbot.pl -i
@{#! @<Perl directory@>

    @<Perl language modes@>
    use LWP;
    
    my $browser = LWP::UserAgent->new();
    $browser->env_proxy();
    
    my $verbose = 0;
    my $settings = '';

command:
    for (my $n = 0; $n <= $#ARGV; $n++) {
        my $cmd = $ARGV[$n];
        
        if ($cmd !~ m/^\-/) {
            $cmd = '--get';
            $n--;
        }
        
        if ($cmd =~ m/^\-/) {
            my $opt = $cmd;
            $opt =~ s/\-+//;
#print("$n: command $opt\n");

            @<Agent command handler@>
            @<Get command handler@>
            @<Head command handler@>
            @<Help command handler@>
            @<Loop command handler@>
            @<Post command handler@>
            @<Set command handler@>
            @<Wait command handler@>
            @<Verbose command handler@>
            
            die("Unknown command $cmd");
        }
    }
        
@}
}

\vbox{
\section{Command handlers}

The following sections handle the various command line options
(which are actually command, not modal options).  Each tests
for the option, then does whatever action it requests.  If the
option takes an argument, it is scanned and validated within
the option handling code.  Note that each command handler must
end with ``{\tt next~command}'' to avoid falling through into
the unknown command code.
}

\vbox{
\subsection{URL Fetch commands}

The following macro retrieves the URL in \verb+$arg+ with the
protocol given in the argument.  If the user has requested
one or more header variables be set with the {\tt --set} command,
we build the fetch method call as a string and execute it
with {\tt eval}, making a special case for POST requests,
which require a dummy form data array argument before the list
of header variables.  It's ugly, but it gets you there.

@d URL Fetch command
@{
    if ($verbose) {
        my $c = "@1";
        $c =~ tr/a-z/A-Z/;
        print("Fetching $arg with $c protocol.\n");
    }
    my $reply;
    
    if ($settings eq '') {
        $reply = $browser->@1($arg);
    } else {
        my @@pform;
        my $parg = ("@1" eq 'post') ? ', \@@pform' : '';
        my $s = '$reply = $browser->@1($arg' . $settings . ')';
        eval($s);        
    }
    if ($verbose) {
        printf("%s:  Status: %s  Content length: %d\n",
            ($reply->is_success ? "Success" : "Failure"),
            $reply->status_line, length($reply->content));
    }
    next command;
@}
}

\vbox{
\subsubsection{{\tt --get} {\em url}}

The {\tt --get} command retrieves the specified {\em url} with
the ``{\tt GET}'' protocol.

@d Get command handler
@{
    if ($opt =~ m/^g/) {
        @<Get command argument@>
        @<URL Fetch command@(get@)@>
    }
@}
}

\vbox{
\subsubsection{{\tt --head} {\em url}}

The {\tt --head} command retrieves the specified {\em url} with
the ``{\tt HEAD}'' protocol.

@d Head command handler
@{
    if ($opt =~ m/^hea/) {
        @<Get command argument@>
        @<URL Fetch command@(head@)@>
    }
@}
}

\vbox{
\subsubsection{{\tt --post} {\em url}}

The {\tt --post} command retrieves the specified {\em url} with
the ``{\tt POST}'' protocol.

@d Post command handler
@{
    if ($opt =~ m/^p/) {
        @<Get command argument@>
        @<URL Fetch command@(post@)@>
    }
@}
}

\vbox{
\subsection{{\tt --agent} {\em agent\_name}}

The {\tt --agent} command sets the {\tt User-Agent} field for
requests.

@d Agent command handler
@{
    if ($opt =~ m/^a/) {
        @<Get command argument@>
        $browser->agent($arg);
        if ($verbose) {
            print("User-Agent set to ", $browser->agent(), "\n");
        }
        next command;
    }
@}
}

\vbox{
\subsection{{\tt --help}}

The {\tt --wait} command prints how-to-call information.
@d Help command handler
@{
    if ($opt =~ m/^hel/) {
        @<Echo command without argument@>
        print << 'EOD';
Usage: badbot.pl [ commands ]
       Commands:
             --agent name       Set User-Agent to name
             --get url          Retrieve url with GET protocol
             --head url         Retrieve url header with HEAD protocol
             --help             Print this message
             --loop             Loop executing commands
             --post url         Retrieve url with POST protocol
             --set item=value   Add item to request header with specified value
             --verbose          Print verbose debugging information
             --wait n           Sleep for n seconds
        A command which doesn't begin with a "-" is submitted as a
        URL with GET protocol.
Version @<Version@>, @<Release Date@>
EOD
        next command;
    }
@}
}

\vbox{
\subsection{{\tt --loop}}

The {\tt --loop} command, which must be last on the command line,
causes BadBot to loop forever over the previous commands.

@d Loop command handler
@{
    if ($opt =~ m/^l/) {
        @<Echo command without argument@>
        if ($n != $#ARGV) {
            die("The --loop command must be the last on the command line");
        }
        $n = -1;
        next command;
    }
@}
}

\vbox{
\subsection{{\tt --set} {\em item}{\tt =}{\em value}}

The {\tt --set} command causes a header variable to be included
in the HTTP request with the specified {\em item} name and
{\em value}.  Any number of {\tt --set} commands may be specified;
each supplying a different header field to the request.  To keep
{\tt --set} commands from accreting without bound when commands
are executed in a {\tt --loop}, they are removed from the command
line after being processed.

@d Set command handler
@{
    if ($opt =~ m/^s/) {
        @<Get command argument@>
        if ($arg !~ m/^([^=]+)=(.*)$/) {
            die("Syntax error in $cmd option argument");
        }
        $settings .= ", '$1' => '$2'";
#print("Before N = $n ", join(', ', @@ARGV), "\n");
        splice(@@ARGV, $n - 1, 2);
        $n -= 2;
#print("After N = $n ", join(', ', @@ARGV), "\n");
        if ($verbose) {
            print("Request settings: $settings\n");
        }
        next command;
    }
@}
}

\vbox{
\subsection{{\tt --verbose}}

The {\tt --verbose} command sets a flag which causes progress and
status messages to be printed on standard output as commands
are executed.

@d Verbose command handler
@{
    if ($opt =~ m/^v/) {
        $verbose = 1;
        @<Echo command without argument@>
        next command;
    }
@}
}

\vbox{
\subsection{{\tt --wait} {\em seconds}}

The {\tt --wait} command causes the program to sleep for the
requested number of {\em seconds}.

@d Wait command handler
@{
    if ($opt =~ m/^w/) {
        @<Get command argument@>
        
        if ($arg !~ m/^\d+$/) {
            die("Argument to $cmd must be numeric.\n");
        }
        if ($verbose) {
            print("Waiting $arg seconds.\n");
        }
        sleep($arg);
        next command;
    }
@}
}

\vbox{
\subsection{Get command argument}

The next command line argument is taken and stored into
\verb+$arg+.  If no argument is specified, an error message
is issued and we give up.

@d Get command argument
@{
    if ($n == $#ARGV) {
        die("Argument missing after $cmd command");
    }
    my $arg = $ARGV[++$n];
    @<Echo command with argument@>
@}
}

\vbox{
\subsection{Echo command without argument}

If {\tt --verbose} is set, this code echoes the current
command (which has no argument) to standard output.

@d Echo command without argument
@{
    if ($verbose) {
        printf("%3d: %s\n", $n, $cmd);
    }
@}
}

\vbox{
\subsection{Echo command with argument}

If {\tt --verbose } is set, this code echoes the current
command and its argument to standard output.

@d Echo command with argument
@{
    if ($verbose) {
        printf("%3d: %s %s\n", $n - 1, $cmd, $arg);
    }
@}
}

\vbox{
\chapter{{\tt Makefile}}

This is the {\tt Makefile} for Gardol.  Of
course, generating the {\tt Makefile} from the Nuweb invites
infinite regress, since it's the {\tt Makefile} which invokes
{\tt nuweb} to create$\ldots$.  But as long as we include the
generated {\tt Makefile} in the source distribution, all will
be well, and we do that below, in the definition of the
{\tt Makefile} in the Nuweb.  {\bf Slap!}  Thanks---I needed
that.

Since, in the interest of preserving formatting in the
\LaTeX\ code documentation, we edit this file with hardware
tabs disabled, we must cope with the regrettable detail
that {\tt make} uses tabs as a significant character.

@o Makefile.mkf -i
@{

WEBDIR = @<Web Directory@>

LOGDIR = @<Web Log File Directory@>

PROGRAMS = gardol.pl

duh:
        @@echo "Please choose: check dist publish test weblint"

@<Extract source code from Nuweb@>

@<Installation@>        

@<Source distribution@>        
        
@<Documentation@>

@<Testing@>
@}
}

\vbox{
\section{Extract source code from Nuweb}

All of the source code for Gardol, its support files, documentation,
and the tools used to build it are defined in the Nuweb file
{\tt gardol.w}.  Processing this file with {\tt nuweb} suffices
to extract all the contents, so we can use the Perl source code
{\tt gardol.pl} as a proxy for all the files generated from the
Nuweb program.  Any {\tt Makefile} target which requires a file
from the Nuweb can simply specify {\tt gardol.pl} as a dependency
and be sure everything is up to date.

One little detail\ldots since the {\tt Makefile} itself is defined
here, when you make a change you must first do something that
processes the Nuweb (``{\tt make~check}'' is a good choice) before
the {\tt Makefile} will contain the changes you made.

@d Extract source code from Nuweb
@{   
gardol.pl:    gardol.w
        nuweb gardol
        chmod 755 $(PROGRAMS)
        unexpand -a <Makefile.mkf >Makefile
@}
}

\vbox{
\section{Source distribution}

@d Source distribution
@{   
dist:   $(PROGRAMS) pdf
        rm -f gardol.tar gardol.tar.gz
        tar cfv gardol.tar gardol.w gardol.pl Makefile ipf_conf_template.txt index.html gardol.pdf \
                drop.sty badbot.pl
        gzip gardol.tar
@}
}

\vbox{
\section{Documentation}

@d Documentation
@{   
view:   gardol.pl
        latex gardol
        nuweb gardol
        latex gardol
        xdvi -s 0 gardol
        
viewman: gardol.pl
        pod2man gardol.pl >ZZgardol.1
        groff -X -man ZZgardol.1
        rm -f ZZgardol.1
        
pdf:    gardol.pl
        sed 's/%%%PDF%%%//' <gardol.tex >ZZgardol.tex
        pdflatex ZZgardol
        pdflatex ZZgardol
        mv ZZgardol.pdf gardol.pdf
        rm -f ZZgardol*
        
viewpdf: pdf
        acroread gardol.pdf
@}
}

\vbox{
\section{Testing}

@d Testing
@{           
check:  $(PROGRAMS)
        perl -c gardol.pl
        perl -c badbot.pl
        weblint index.html
        
test:   $(PROGRAMS)
        perl gardol.pl --verbose $(LOGDIR)/forensic_log $(LOGDIR)/ipfilter_log
        
weblint:    gardol.pl
        weblint index.html
@}
}

\vbox{
\section{Installation}

@d Installation
@{   
publish:    dist
        cp -p index.html gardol.tar.gz gardol.pdf $(WEBDIR)
        cp -p figures/* $(WEBDIR)/figures
@}
}

\vbox{
\chapter{Indices} \label{indices}

Three sets of indices can be created automatically: an index of file
names, an index of macro names, and an index of user-specified
identifiers. An index entry includes the name of the entry, where it
was defined, and where it was referenced.
}

\section{Files}

@f

\section{Macros}

@m

\section{Identifiers}

Sections which define identifiers are underlined.

@u

\chapter{Development Log} \label{log}

\font\df=cmbx12
\def\date#1{{\medskip\noindent\df #1\medskip}}
\parskip=1ex
\parindent=0pt

\date{2004 March 1}

Created Nuweb {\tt gardol.w} from initial hacked version of
Perl program.

\date{2004 March 3}

Added command line options to override all internal configuration variable
settings.

\date{2004 March 5}

Removed hard-coded pattern definition for HTTP log parsing.  The patterns
for parsing common, combined, and forensic log formats are now defined as
regular expression variables, and a configuration variable,
\verb+$parseHTTPlog+, is set to the pattern to be used.  Similarly,
the pattern used to parse the IP Filter log is now defined by
a variable \verb+$parse_IP_Filter+.

Integrated logo for Web page and first cut of section headings for the
Web document.

Debugged distribution archive creation and the ``{\tt publish}'' target
in the {\tt Makefile} which creates the Web tree.

Integrated manual-page template documentation in POD format, included
at the head of the emitted {\tt gardol.pl} file.

\date{2004 March 10}

Added a {\tt --blacklist} facility, which permits specifying a file
containing IP addresses (or more general IP Filter ``{\tt from}'' expressions)
which are unconditionally copied into the block list.  The blacklist
is re-read from the file every time the filter list is updated, so
it can be changed at any time.  An error is reported if the
{\tt --blacklist} option is specified and the blacklist file
does not exist, an error is reported, but should the blacklist file
not exist at the time of an update, it will simply be ignored.
This allows users to rename blacklist files to switch them on
and off and/or swap blacklists without causing an error.

\date{2004 March 12}

Added code to catch the {\tt HUP} signal and, after receiving it, the
next time around the main monitoring loop close, re-open, and re-seek
the log files to the end.  This allows following log files are they
are cycled by renaming and restarting the programs that write them.

The current build is now in production on Vitesse.  It's run from a
shell script which specifies all the file locations.  There is
no need for \verb+$prime_time+ to override built-in path name
defaults, so it has been removed.

Added the ability to transcribe arbitrary IP Filter rules from the
{\tt --blacklist} file to the filter definition.  Any line whose
first nonblank character is a plus sign is transcribed explicitly,
dropping the plus sign.  No syntax or other checking is performed--the
text is simply copied into the filter configuration file.

Changed the default for {\tt --ipfupdate} to the null string,
disabling automatic update.  It's better not to have reloading
the IP filter a built-in default.

\date{2004 March 13}

The command line option documentation in the POD document was
getting unwieldy, so I broke it up into one section per
option.  This results in a nice list of options in the table
of contents, which will be linked in the PDF.

Broke the {\tt Makefile} up into logical sections of targets
for build, installation, documentation, test, etc.  Added an
explanation to the build target about how we use {\tt gardol.pl}
as a proxy for all files generated from {\tt gardol.w} and the
quirk about having to make twice before a change in the
{\tt Makefile} takes effect.

The ``{\tt --Timeout}'' diagnostic was printed on standard
error unconditionally.  Since it is redundant given the
``{\tt --Timeout purged}'' message issued in {\tt --verbose}
mode, I just removed it.

\date{2004 March 13}

Added ``{\tt -}'' as a default argument for the {\tt --ipfupdate},
{\tt --ipfconf}, {\tt --ipftemplate}, and {\tt --ipfupdate} options.
The default argument selects the file names configured as
Operating System Default Parameters (\ref{defargs}).

\date{2004 March 15}

Added the BadBot attack simulator program.  Cleanup and more documentation
to follow, as always.

\end{document}
